# Generated by Django 4.2.4 on 2023-08-14 10:35

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Budget',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('description', models.TextField(blank=True, null=True)),
                ('active', models.BooleanField(default=True)),
            ],
        ),
        migrations.CreateModel(
            name='Payee',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('description', models.TextField(blank=True, null=True)),
                ('budget', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='budgetmanager.budget')),
            ],
        ),
        migrations.CreateModel(
            name='Payment',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', models.DecimalField(decimal_places=2, max_digits=7)),
                ('date', models.DateField()),
                ('pending', models.BooleanField(default=False, verbose_name='Exclude from total')),
                ('notes', models.TextField(blank=True, null=True)),
                ('payee', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='budgetmanager.payee')),
            ],
        ),
        migrations.CreateModel(
            name='BudgetShare',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('can_edit', models.BooleanField(default=False)),
                ('budget', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='budgetmanager.budget')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AddField(
            model_name='budget',
            name='shared_users',
            field=models.ManyToManyField(blank=True, related_name='shared_budgets', through='budgetmanager.BudgetShare', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='budget',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddConstraint(
            model_name='budgetshare',
            constraint=models.UniqueConstraint(fields=('budget', 'user'), name='One budget membership per user per budget'),
        ),
    ]



'''
URL configuration for budgetmanager app
'''
from django.urls import path
from rest_framework.routers import DefaultRouter
from . import views

app_name = 'budgetmanager'  # pylint: disable=invalid-name
router = DefaultRouter()
router.register(r'budget', views.BudgetViewSet)
router.register(r'share', views.BudgetShareViewSet)
router.register(r'payee', views.PayeeViewSet)
router.register(r'payment', views.PaymentViewSet)

urlpatterns = router.urls + [
    path('total/', views.TotalView.as_view()),
]



'''
Model classes
'''
# pylint:disable=no-member
from datetime import datetime
from decimal import Decimal

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models


def _get_total_amount(queryset) -> Decimal:
    return queryset.filter(pending=False).aggregate(
        models.Sum('amount', default=0)
    )['amount__sum']


class Budget(models.Model):
    '''
    Model for a budget
    '''
    user = models.ForeignKey(settings.AUTH_USER_MODEL,
                             on_delete=models.CASCADE)
    name = models.CharField(max_length=100)
    description = models.TextField(null=True, blank=True)
    active = models.BooleanField(default=True)
    shared_users = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        through='BudgetShare',
        related_name='shared_budgets',
        blank=True
    )

    def __str__(self):
        return str(self.name)

    def add_from_csv(self, text: str):
        '''
        Add payees and payments to this budget from a CSV formatted string
        '''
        rows = text.strip().split('\n')
        for line in rows:
            record = line.split(',')
            payee = Payee.objects.get_or_create(
                name=record[0],
                budget=self
            )[0]
            payment = Payment(
                payee=payee,
                amount=record[1],
                date=datetime.strptime(record[2], '%d/%m/%Y'),
            )
            if len(record) >= 4:
                payment.notes = record[3]
            if len(record) >= 5:
                payment.pending = record[4] != ''
            payment.save()

    @property
    def total(self):
        '''The total amount of the Payments of this Budget'''
        return _get_total_amount(Payment.objects.filter(payee__budget=self))

    @classmethod
    def get_user_model(cls):
        return cls._meta.get_field('user').related_model


class BudgetShare(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL,
                             on_delete=models.CASCADE)
    budget = models.ForeignKey(Budget, on_delete=models.CASCADE)
    can_edit = models.BooleanField(default=False)

    def clean(self):
        if self.user == self.budget.user:
            raise ValidationError('Budget owner cannot be a shared user')

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=('budget', 'user'), name='One budget membership per user per budget'),
        ]


class Payee(models.Model):
    '''
    Model for a payee
    '''
    budget = models.ForeignKey(Budget, on_delete=models.CASCADE)
    name = models.CharField(max_length=100)
    description = models.TextField(null=True, blank=True)

    def __str__(self):
        return str(self.name)

    @property
    def total(self):
        '''
        The total amount of the Payments of this Payee
        '''
        return _get_total_amount(self.payment_set)


class Payment(models.Model):
    '''
    Model for a payment
    Requires a payee and a budget
    Has an amount and date
    '''
    payee = models.ForeignKey(
        Payee,
        on_delete=models.CASCADE,
    )
    amount = models.DecimalField(decimal_places=2, max_digits=7)
    date = models.DateField()
    pending = models.BooleanField(
        default=False, verbose_name='Exclude from total')
    notes = models.TextField(null=True, blank=True)

    @classmethod
    def get_total(cls, user):
        '''Get the total amount of the user's Payments'''
        return _get_total_amount(cls.objects.filter(payee_budget__user=user))



# pylint: disable=unused-argument
from rest_framework.permissions import BasePermission


class IsBudgetOwner(BasePermission):
    def has_obj_permission(self, request, view, obj):
        return obj.user == request.user


class CanAccessBudgetShare(BasePermission):
    def has_object_permission(self, request, view, obj):
        return obj.user == request.user or obj.budget.user == request.user


class IsPayeeOwner(BasePermission):
    def has_obj_permission(self, request, view, obj):
        return obj.budget.user == request.user


class IsPaymentOwner(BasePermission):
    def has_obj_permission(self, request, view, obj):
        return obj.payee.budget.user == request.user



from rest_framework.serializers import (
    ModelSerializer,
    HiddenField,
    CurrentUserDefault
)

from . import models


class UserSerializer(ModelSerializer):
    class Meta:
        model = models.Budget.get_user_model()
        fields = ('username', 'first_name', 'last_name')


class BudgetSerializer(ModelSerializer):
    user = HiddenField(default=CurrentUserDefault())

    class Meta:
        model = models.Budget
        fields = ('id', 'name', 'description', 'active', 'user')


class BudgetShareSerializer(ModelSerializer):
    user = UserSerializer(read_only=True)

    class Meta:
        model = models.BudgetShare
        fields = ('id', 'budget', 'user')


class PayeeSerializer(ModelSerializer):
    class Meta:
        model = models.Payee
        fields = ('id', 'name', 'description', 'budget')


class PaymentSerializer(ModelSerializer):
    class Meta:
        model = models.Payment
        fields = ('id', 'notes', 'payee', 'amount', 'date', 'pending')



'''Admin setup for budgetmanager app'''
from django.contrib import admin
from . import models


@admin.register(models.Budget)
class BudgetAdmin(admin.ModelAdmin):
    '''Settings for the Budget admin'''
    list_display = ('user', 'name', 'active')
    list_display_links = list_display
    list_filter = ('user', 'active')
    sortable_by = list_display
    search_fields = ('name',)
    search_help_text = 'Search by budget name'
    list_per_page = 20


@admin.register(models.BudgetShare)
class BudgetShareAdmin(admin.ModelAdmin):
    '''Settings for the BudgetShare admin'''
    list_display = ('user', 'budget', 'can_edit')
    list_display_links = list_display
    list_filter = list_display
    sortable_by = list_display
    readonly_fields = ('user', 'budget')
    list_per_page = 20

    def get_readonly_fields(self, request, obj=...):
        if obj is None:
            return tuple()
        return self.readonly_fields


@admin.register(models.Payee)
class PayeeAdmin(admin.ModelAdmin):
    '''Settings form the Payee admin'''
    list_display = ('budget', 'name')
    list_display_links = list_display
    list_filter = ('budget',)
    sortable_by = list_display
    search_fields = ('name',)
    search_help_text = 'Search by payee name'
    list_per_page = 20


@admin.register(models.Payment)
class PaymentAdmin(admin.ModelAdmin):
    '''Settings for the Payment admin'''
    date_hierarchy = 'date'
    list_display = ('payee', 'date')
    list_display_links = list_display
    list_filter = ('payee', 'date')
    sortable_by = list_display
    list_per_page = 20



# pylint: disable=no-member
from django.db.models import Q
from rest_framework import mixins, status
from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.viewsets import GenericViewSet, ModelViewSet

from . import (
    models,
    permissions,
    serializers,
)
from .pagination import Pagination


class TotalView(APIView):
    permission_classes = (IsAuthenticated,)

    def get(self, request):
        return Response(
            f'{models.Payment.get_total(request.user):.2f}'
        )


class PaymentRelatedMixin(ModelViewSet):
    @action(methods=('GET',), detail=True)
    def total(self, request, pk):
        return Response({
            'total': f'{self.get_object().total:.2f}'
        })


class BudgetViewSet(PaymentRelatedMixin, ModelViewSet):
    queryset = models.Budget.objects
    serializer_class = serializers.BudgetSerializer
    permission_classes = (IsAuthenticated, permissions.IsBudgetOwner)
    pagination_class = Pagination

    def get_queryset(self):
        return self.queryset.filter(user=self.request.user).all()

    @action(methods=('POST',), detail=True, url_path='csv')
    def add_from_csv(self, request, pk):
        self.get_object().add_from_csv(request.data['csv'])
        return Response(None, status=status.HTTP_204_NO_CONTENT)


class BudgetShareViewSet(
    mixins.RetrieveModelMixin,
    mixins.DestroyModelMixin,
    mixins.ListModelMixin,
    GenericViewSet
):
    queryset = models.BudgetShare.objects
    serializer_class = serializers.BudgetShareSerializer
    permission_classes = (IsAuthenticated, permissions.CanAccessBudgetShare)
    pagination_class = Pagination

    def get_queryset(self):
        return self.queryset.filter(
            Q(user=self.request.user)
            | Q(budget__user=self.request.user)
        ).all()


class PayeeViewSet(PaymentRelatedMixin, ModelViewSet):
    queryset = models.Payee.objects
    serializer_class = serializers.PayeeSerializer
    permission_classes = (IsAuthenticated, permissions.IsPayeeOwner)
    pagination_class = Pagination

    def get_queryset(self):
        return self.queryset.filter(budget__user=self.request.user).all()


class PaymentViewSet(ModelViewSet):
    queryset = models.Payment.objects
    serializer_class = serializers.PaymentSerializer
    permission_classes = (IsAuthenticated, permissions.IsPaymentOwner)
    pagination_class = Pagination

    def get_queryset(self):
        return self.queryset.filter(payee__budget__user=self.request.user).all()
